! ============================================================================================================================
! MODULE        : stomate_pest
!
! CONTACT       : orchidee-help _at_ ipsl.listes.ipsl.fr
!
! LICENCE       : IPSL (2006).
! This software is governed by the CeCILL licence see ORCHIDEE/ORCHIDEE_CeCILL.LIC
!
!>\BRIEF        : Calculates tree damage due to bark-beetle outbreaks 
!!
!!\n DESCRIPTION: See subroutine bark_beetle_damage for a more elaborate description of the
!!              principles.
!!
!! RECENT CHANGE(S): None
!!
!! REFERENCES: C. TEMPERLI, H. BUGMANN, AND C. ELKIN. Cross-scale interactions
!!             among bark beetles, climate change, and wind disturbances: a
!!             landscape modeling approach. Ecological Monographs, 83(3), 2013, pp. 383–402. 
!!
!! SVN           :
!! $HeadURL: svn://forge.ipsl.jussieu.fr/orchidee/branches/ORCHIDEE-DOFOCO/ORCHIDEE/src_stomate/stomate_pest.f90 
!! $Date: 2019-01-22 13:41:13 +0100 (mar., 06 aout. 2019)
!! $Revision: $
!! \n
!_ ===========================================================================================================================

MODULE stomate_pest

  ! modules used:
  USE ioipsl_para
  USE xios_orchidee
  USE stomate_data
  USE constantes
  USE constantes_soil
  USE pft_parameters
  USE pft_parameters_var
  USE function_library, ONLY: wood_to_ba, wood_to_qmdia, Nmax, wood_to_dia
  IMPLICIT NONE

  ! private & public routines

  PRIVATE
  PUBLIC  pest_clear,bark_beetle_damage

  LOGICAL, SAVE                   :: firstcall_bark_beetle_damage = .TRUE.     !! First call 
!$OMP THREADPRIVATE(firstcall_bark_beetle_damage)
  INTEGER(i_std), SAVE            :: printlev_loc                       !! Local level of text output
!$OMP THREADPRIVATE(printlev_loc)

CONTAINS

!!
!==================================================================================================================================
!!  SUBROUTINE   : pest_clear 
!!
!>\BRIEF        Set the firstcall flag to .TRUE. and activate initialization
!! 
!_
!================================================================================================================================

  SUBROUTINE pest_clear
    firstcall_bark_beetle_damage = .TRUE.
  END SUBROUTINE pest_clear


!================================================================================================================================
!! SUBROUTINE    : bark_beetle_damage
!!
!>\BRIEF         Calculates tree damage due to bark-beetle outbreaks 
!!
!!\n DESCRIPTION : This subroutine is used to calculate the annual amount of trees kill by
!!              bark beetles. This module is strongly inspired by the work of Temperli
!!              et al. 2013 in Ecological Monographs. The bark beetle module developped 
!!              for LANDCLIM landscape model was the basis of this work. The module was
!!              adapted to fit the constraints of ORCHIDEE in regard to its spatial and
!!              temporal resolution. In short, this module calculate a suceptibility of
!!              of a forest to be attack by bark beetles at the pixel/PFT level. At 
!!              present the model has been parameterized for Picea abies age. The risk
!!              probability of a pixel to be attacked depends on pixel suceptibility and
!!              bark beetle presure where pressure is calculated as an index that
!!              indicates the activity of the beetles in this pixel). Then, a mortality
!!              rate for each Picea abies PFT is calculated based on PFT suceptibility
!!              and beetle pressure. Finally, the wood biomass kill by bark beetle of a
!!              given PFT is the product of the pixel risk * mortality rate * the PFT
!!              wood biomass. For each PFT we first kill the small trees and if more
!!              biomass needs to be killed, ORCHIDEE will continue with killing the
!!              larger diameter classes.
!!              The most important difference beetween this module and the original one
!!              in LANDCLIM is that ORCHIDEE lost all information about the spatialization
!!              of the attack. In other words the attack is no longer spatialy explicit.
!!  
!! RECENT CHANGE(S): Added in December 2019
!!
!! MAIN OUTPUT VARIABLE(S) : circ_class_kill
!!
!! REFERENCES: CHRISTIAN TEMPERLI, HARALD BUGMANN, AND CHE ELKIN. Cross-scale
!!             interactions among bark beetles, climate change, and wind
!!             disturbances: a landscape modeling approach. Ecological Monographs,
!!             83(3), 2013, pp. 383–402.
!!
!! FLOWCHART     : 
!! \n
!_================================================================================================================================

  SUBROUTINE bark_beetle_damage( &
                npts,  circ_class_biomass, beetle_generation_index,&
                veget_max, circ_class_n, circ_class_dist, &
                age_stand, season_drought_legacy, resolution, &
                circ_class_kill, forest_managed, wood_leftover_legacy, &
                beetle_pop_legacy, risk_index_legacy, & 
                beetle_damage_legacy, risk_index, epidemic_monitor, &
                epidemic, beetle_flyaway, windthrow_suscept_monitor, &
                beetle_pressure_monitor, suscept_index_monitor )

    IMPLICIT NONE
   
    !! 0. Variable declarations
    !! 0.1 Input variables
    INTEGER(i_std), INTENT(in)                         :: npts                     !! number of pixels
    INTEGER(i_std), DIMENSION (:, :), INTENT(in)       :: forest_managed           !! forest management flag 
                                                                                   !! (0 : unmanaged, >0 : managed)
    REAL(r_std), DIMENSION(:, :, :, :, :), INTENT(in)  :: circ_class_biomass       !! Biomass of an individual tree (gC.m-2) 
                                                                                   !! in each circ class
    REAL(r_std), DIMENSION(:, :, :), INTENT(in)        :: circ_class_n             !! Number of trees(tree.m-2) 
                                                                                   !! within each circumference
    REAL(r_std), DIMENSION(:), INTENT(in)              :: circ_class_dist          !! The probability distribution of trees
    REAL(r_std), DIMENSION(:, :), INTENT(in)           :: veget_max                !! "maximal" coverage fraction of 
                                                                                   !! a PFT on the ground
    REAL(r_std), DIMENSION(:, :), INTENT(in)           :: resolution               !! The length in m of each grid-box in 
                                                                                   !! X and Y axis
    REAL(r_std), DIMENSION(:, :, :), INTENT(in)        :: beetle_generation_index  !! number of generation that BB can 
                                                                                   !! achieved in one year
    INTEGER(i_std), DIMENSION(:, :), INTENT(in)        :: age_stand                !! Age of stand (years)
    REAL(r_std), DIMENSION(:, :, :), INTENT(in)        :: season_drought_legacy    !! mean growing season moisture availability 
    REAL(r_std), DIMENSION(:, :, :), INTENT(in)        :: wood_leftover_legacy     !! Coarse woody bedris, snags and logs
                                                                                   !! (gC.m-2) left on the PFT 
                                                                                   !! during previous timestep
    REAL(r_std), DIMENSION(:, :, :), INTENT(in)        :: risk_index_legacy        !! risk index stored every year, n time 
    !! 0.3 Modified variables
    REAL(r_std), DIMENSION(:, :, :, :, :), INTENT(inout) :: circ_class_kill        !! Number of trees within a circ that needs
    REAL(r_std), DIMENSION(:, :, :),INTENT(inout)        :: beetle_pop_legacy      !! Index (0-1) Biomass of tree from the 
                                                                                   !! same PFT that was
    REAL(r_std), DIMENSION(:, :), INTENT(inout)        :: epidemic                 !! logical flag indicationg if beetle
                                                                                   !! outbreak is in a epidemic stage
    REAL(r_std), DIMENSION(:, :),INTENT(inout)         :: risk_index               !! risk that a given pixel experience an 
                                                                                   !! outbreak (unitless)
    REAL(r_std), DIMENSION(:, :), INTENT(inout)        :: beetle_flyaway
    REAL(r_std), DIMENSION(:, :), INTENT(inout)        :: epidemic_monitor         !! variable used to monitor epidemic flag
    REAL(r_std), DIMENSION(:, :, :), INTENT(inout)     :: beetle_damage_legacy     !! temporary variable(gC.m-2)
                                                                                   !! infected during the previous timestep  
    REAL(r_std), DIMENSION(:, :), INTENT(inout)        :: windthrow_suscept_monitor!! variable used to monitor windthrow susceptibility (unitless)
    REAL(r_std), DIMENSION(:, :), INTENT(inout)        :: beetle_pressure_monitor  !! variable used to monitor beetle pressure index (unitless)
    REAL(r_std), DIMENSION(:, :), INTENT(inout)        :: suscept_index_monitor    !! variable used to monitor overall susceptibility index (unitless)

    !! 0.4 Local variables
    INTEGER(i_std)                                    :: ipts, ileg, ivm, icir, &
                                                         ivmap, iagec, iyear, xvm, &
                                                         ncirc_loc
    INTEGER(i_std), DIMENSION(npts, nvmap, nagec)     :: nc                        !! number of PFT counter
                                                                                   !! within a species group that are
                                                                                   !! infected by bark beetles (unitless) 
    REAL(r_std), DIMENSION(npts, nvm)                 :: season_drought_mean       !! average drought index on this PFT 
                                                                                   !! during previous time steps.
    REAL(r_std), DIMENSION(npts, nvm)                 :: wood_leftover_mean        !! average biomass (gC m-2)
                                                                                   !! of dead wood left on this PFT
                                                                                   !! during previous time steps.
    REAL(r_std), DIMENSION(npts, nvmap)               :: windthrow_suscept_gp      !! Species group-level susceptibility 
                                                                                   !! to bark beetle attacks due to the 
                                                                                   !! leftover of windthrow debris (unitless)
    REAL(r_std), DIMENSION(npts, nvmap)               :: drought_suscept_gp        !! Species group-level susceptibility 
                                                                                   !! to bark beetle attacks due to weakening 
                                                                                   !! from droughts (unitless)
    REAL(r_std), DIMENSION(npts, nvmap)               :: standage_suscept_gp       !! Species group-level susceptibility 
                                                                                   !! to bark beetle attacks
    REAL(r_std), DIMENSION(npts, nvmap)               :: rdi_suscept_gp            !! Species group-level susceptibility 
                                                                                   !! to relative density of stem
                                                                                   !! due to the stand age (unitless)
    REAL(r_std), DIMENSION(npts, nvmap)               :: share_suscept_gp          !! Species group-level susceptibility 
                                                                                   !! to bark beetle attacks due to
                                                                                   !! the landscape heterogeneity (unitless)
    REAL(r_std), DIMENSION(npts, nvmap)               :: suscept_index_gp          !! Species group-level susceptibility 
                                                                                   !! to bark beetle attacks due to 
                                                                                   !! the combined effect of windthrow, 
                                                                                   !! stand age and
    REAL(r_std), DIMENSION(npts, nvmap)               :: risk_index_gp             !! risk index averaged for all
                                                                                   !! age classes of the same PFT
                                                                                   !! droughts (unitless)
    REAL(r_std), DIMENSION(npts, nvmap)               :: beetle_pressure_index     !! Beetle activity of a given pixel
    REAL(r_std), DIMENSION(npts, nvmap)               :: biomass_infected          !! Total biomass of a species group 
                                                                                   !! that is infected by bark
                                                                                   !! beetles (gC or gN m-2). 
    REAL(r_std), DIMENSION(npts, nvmap)               :: vegetmax_gp               !! share of all PFTs of the same species 
                                                                                   !! (called species group) within
    REAL(r_std), DIMENSION(npts)                      :: vegetmax_deciduous        !! share of deciduous PFTs 
                                                                                   !! a pixel (0-1, unitless)
    REAL(r_std), DIMENSION(npts, nvm)                 :: standage_suscept          !! Pft-level susceptibility
                                                                                   !! due to stand age (0-1, unitless)
    REAL(r_std), DIMENSION(npts, nvm)                 :: drought_suscept           !! Pft-level susceptibility
                                                                                   !! due to weakening from droughts 
                                                                                   !! (0-1, unitless)
    REAL(r_std), DIMENSION(npts, nvm)                 :: suscept_index             !! Pft-level susceptibility to bark beetle 
                                                                                   !! attacks due to the combined effect of 
                                                                                   !! windthrow, stand age and
                                                                                   !! droughts (0-1, unitless)
    REAL(r_std), DIMENSION(npts, nvm)                 :: mortality_rate            !! probability that an infected tree die
    REAL(r_std), DIMENSION(npts, nvm)                 :: beetle_damage             !! bark beetle damage in gC m-2 year-1
    REAL(r_std), DIMENSION(npts, nvmap)               :: G                         !! Average bark beetle pressure during 
                                                                                   !! the three recent years
    REAL(r_std)                                       :: wght_sirdi                !! weigth rdi_susceptibility in the calculation
                                                                                   !! of susceptibility_index 
    REAL(r_std)                                       :: wght_siw                  !! weigth windthrow_susceptibility in the 
                                                                                   !! calculation of susceptibility_index
    REAL(r_std)                                       :: qm_dia                    !! quadratic diameter
    REAL(r_std), DIMENSION(npts, nvmap)               :: woody_frac                !! woody/total biomass (unitless) 
    REAL(r_std), DIMENSION(ncirc)                     :: qdia                      !! quadratic diameter per circumference class
    REAL(r_std), DIMENSION(npts, nvm)                 :: rdi                       !! relative density index
    REAL(r_std), DIMENSION(npts, nvmap)               :: rdi_gp                    !! rdi for all age-classe of a same PFT
    REAL(r_std), DIMENSION(npts, nvmap)               :: epidemic_risk             !! with the highest bark beetle pressure 
                                                                                   !! (0-1, unitless)
    REAL(r_std), DIMENSION(legacy_years)              :: bgi_current               !! temporary variable
    REAL(r_std), DIMENSION(npts, nvmap)               :: biomass_current           !! Actual biomass of the PFT (gC or gN m-2)
    REAL(r_std), DIMENSION(npts, nvmap)               :: woody_current             !! Actual aboveground woody biomass of the PFT 
                                                                                   !! (gC or gN m-2)
    REAL(r_std)                                       :: share                     !! Deciduous/coniferous share in a pixel
    REAL(r_std), DIMENSION(npts, nvm, legacy_years)   :: beetle_temp1               !! temporary variable
    REAL(r_std), DIMENSION(npts, nvm, beetle_legacy)  :: beetle_temp2
    REAL(r_std), DIMENSION(nvm)                       :: rdi_limit                 !! Rdi a which epidemic come back to endimic (unitless)
    LOGICAL, DIMENSION(npts, nvmap)                   :: beetle_gp                 !! Flag that indicates whether or not 
                                                                                   !! bark beetle damage should
                                                                                   !! be calculated for that species group. 
!================================================================================================================================

    IF (firstcall_bark_beetle_damage) THEN
       !! Initialize local printlev
       printlev_loc = get_printlev('barkbeetle_damage')
       firstcall_bark_beetle_damage = .FALSE.
    END IF

    IF (printlev_loc .GE. 2) WRITE(numout, *) 'Entering stomate pest damage'
    ! Initialize variables
    vegetmax_gp(:, :) = zero
    standage_suscept_gp(:, :) = zero
    windthrow_suscept_gp(:, :) = zero
    drought_suscept_gp(:, :) = zero
    share_suscept_gp(:, :) = zero
    rdi_suscept_gp(:, :) = zero
    risk_index(:, :) = zero
    rdi(:, :) = zero
    rdi_gp(:, :) = zero
    biomass_current(:, :) = zero
    woody_current(:, :) = zero
    woody_frac(:, :) = un
    season_drought_mean(:, :) = zero
    wood_leftover_mean(:, :) = zero
    beetle_damage(:, :) = zero
    mortality_rate(:, :) = zero
    G(:, :) = zero
    nc(:, :, :) = zero
    ! THe end of an epidemic is trigger by the forest density (RDI) 
    ! rdi_limit must be above rdi_susceptibility_b in order to avoid
    ! long tail in epidemic phase the rdi_target_suscept must be ranged
    ! between 0.5 < rdi_target_suscept < 0.7. 
    rdi_limit(:) = log(1 / rdi_target_suscept(:) - 1) / &
                    rdi_susceptibility_a(:) + &
                    rdi_susceptibility_b(:)

    ! Assume that we are not interested in calculating bark beetle
    ! damage for this species group.
    beetle_gp(:, :) = .FALSE.

    ! This is the main loop for wich the beetle damage will be calculated
    DO ipts = 1, npts

       DO ivmap = 1, nvmap

          ! a fist loop over PFT to setup crusial variable: vegetmax_group,
          ! vegetmax_deciduous, rdi_group, beetle_group and nc
          DO iagec = 1, nagec_pft(ivmap)

             ! Determine the PFT index for the age classes within the species
             ! group under consideration
             ivm = start_index(ivmap) + (iagec - 1)

             WRITE(numout,*) 'PFT, ivmap, ipts', ivm, ivmap, ipts
             IF(printlev_loc >= 4)THEN
                WRITE(numout,*) 'PFT', ivm
                WRITE(numout,*) 'beetle_pft', beetle_pft(ivm)
                WRITE(numout,*) 'beetle_group', beetle_gp(ipts, ivmap)
                WRITE(numout,*) 'veget_max', veget_max(ipts, ivm)
             END IF
             !-

             IF (.NOT. beetle_pft(ivm) .OR. (veget_max(ipts, ivm) .LE. min_stomate)) THEN
                ! We don't want to calculate bark beetle infections for this PFT
                ! or there is no vegetation present in the PFT. nc is a counter.
                ! Because at least one PFT of the age class group has no beetles in it,
                ! the counter is decreased. It will be used later in other  calculations.
                nc(ipts, ivmap, iagec) = 0
             ELSE
                nc(ipts, ivmap, iagec) = 1
                !Relative density index (RDI) is needed here in order to
                !calculate susceptibility to RDI in the next circ_class loop 
                qm_dia = wood_to_qmdia( &
                        circ_class_biomass(ipts, ivm, :, :, icarbon), &
                        circ_class_n(ipts, ivm, :), ivm)

                rdi(ipts, ivm) = (SUM(circ_class_n(ipts, ivm, :)) * m2_to_ha) / &
                        Nmax(qm_dia * m_to_cm, ivm)
                ! Calculate the total area of this species group across all age
                ! classes. Indicate that there bark beetles damage needs to be
                ! calculated in this species group.
                vegetmax_gp(ipts, ivmap) = vegetmax_gp(ipts, ivmap) + veget_max(ipts, ivm)
                beetle_gp(ipts, ivmap) = .TRUE.
             ENDIF

             ! Deciduous tree prevent picea trees to be infested by bark beetle
             ! by acting "smoke screen". We include that effect in the code
             ! later through share susceptibility.
             IF (is_deciduous(ivm) .AND. is_tree(ivm)) THEN
                vegetmax_deciduous(ipts) = vegetmax_deciduous(ipts) + veget_max(ipts, ivm)
             ENDIF

          ENDDO ! iagec

          IF (beetle_gp(ipts, ivmap) .AND. vegetmax_gp(ipts, ivmap) .GT. min_stomate) THEN
            DO iagec = 1, nagec_pft(ivmap)
                IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                  CYCLE
                ENDIF
                ivm = start_index(ivmap) + (iagec - 1)
                ! All picea pft class are conrcerned by suceptibility to RDI
                rdi_gp(ipts, ivmap) = rdi_gp(ipts, ivmap) + rdi(ipts, ivm) * un / SUM(nc(ipts, ivmap, :))

                ! Actual biomass of the PFT (gC or gN m-2)
                biomass_current(ipts, ivmap) = biomass_current(ipts, ivmap) + &
                        SUM(SUM(circ_class_biomass(ipts, ivm, :, :, icarbon), 2) * &
                        circ_class_n(ipts, ivm, :))
                ! Woody biomass of the PFT (gC or gN m-2)
                woody_current(ipts, ivmap) = woody_current(ipts, ivmap) + &
                        SUM((circ_class_biomass(ipts, ivm, :, isapabove,icarbon) + &
                        circ_class_biomass(ipts, ivm, :, iheartabove, icarbon)) * &
                        circ_class_n(ipts, ivm, :))
                woody_frac(ipts, ivmap) = SUM((circ_class_biomass(ipts, ivm, :, isapabove,icarbon) + &
                        circ_class_biomass(ipts, ivm, :, iheartabove, icarbon))* &
                        circ_class_n(ipts, ivm, :)) / &
                        SUM(SUM(circ_class_biomass(ipts, ivm, :, :, icarbon), 2) * &
                        circ_class_n(ipts, ivm, :))
                ! Update the average biomass (gC or gN m-2)
                ! of dead wood left on this PFT during previous time steps.
                ! Only biomass
                ! added during the legacy period is accounted for.
                wood_leftover_mean(ipts, ivmap) = MAX(wood_leftover_mean(ipts, ivmap) + &
                        (SUM(wood_leftover_legacy(ipts, ivm, :)) / legacy_years - &
                         SUM(beetle_damage_legacy(ipts, ivm,:) * &
                         woody_frac(ipts, ivmap))), zero) 
                season_drought_mean(ipts, ivm) = &
                        SUM(season_drought_legacy(ipts, ivm, :)) / legacy_years

            ENDDO 

          ENDIF
 
       ENDDO

       ! The main PFT loop :
       ! Bark beetle damage is calculated across all age classes of the
       ! same species group. Hence the loop over nvmap instead of nvm.
       DO ivmap = 1, nvmap

          ! The fist age class PFT use when PFT parameter is needed in
          ! equation outside iagec loop  
          xvm = start_index(ivmap)

          ! In the original model "share" represent the purity of the stand
          ! (monospecific vs mixte). This index reflect the fact that beetle are
          ! sensitive to deciduous (AGE_GROUP 6 and 8 for 15 PFT) species 
          ! which act as natural "smoke screen".
          IF(vegetmax_deciduous(ipts) .GT. zero) THEN
              DO iagec = 1, nagec_pft(ivmap)
                IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                  CYCLE
                ENDIF
              ivm = start_index(ivmap) + (iagec - 1)
              share = vegetmax_deciduous(ipts) / & 
                   (vegetmax_gp(ipts, ivmap) + vegetmax_deciduous(ipts))
              share_suscept_gp(ipts, ivmap) = &
                   un / (un + exp(share_susceptibility_a(ivm) * &
                   (share - share_susceptibility_b(ivm))))
            ENDDO
          ELSE
            share_suscept_gp(ipts, ivmap) = un
          END IF

          !Debug
          IF(printlev_loc >= 4)THEN
             WRITE(numout,*) 'vegetmax_group', vegetmax_gp(ipts, ivmap)
             WRITE(numout,*) 'rdi_group', rdi_gp(ipts, ivmap)
             WRITE(numout,*) 'vegetmax_deciduous', vegetmax_deciduous(ipts)
             WRITE(numout,*) 'beetle_group', beetle_gp(ipts, ivmap)
             WRITE(numout,*) 'nc', nc(ipts, ivmap, :)
          END IF
          !-
             
          IF (beetle_gp(ipts, ivmap) .AND. vegetmax_gp(ipts, ivmap) .GT. min_stomate) THEN
             ! Bark beetle damage needs to be calculated for this species group and
             ! there is vegetation present.
             ! Epidemic risk is an important concept of this model. It give the
             ! probability that an epidemic phase will happen. An Epidemic state
             ! is define by diferences in the importance of the specific 
             ! susceptibility in the calculation of the infested rate. 
             DO iagec = 1, nagec_pft(ivmap)
               IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                 CYCLE
               ENDIF
               ivm = start_index(ivmap) + (iagec - 1)
               ! risk_index_legacy are averaged is across legacy_year. It is a minimum
               ! and can be increase up to the whole isk_index_legacy vector 
               epidemic_risk(ipts, ivmap) = epidemic_risk(ipts, ivmap) + &
                 SUM(risk_index_legacy(ipts, ivm, :)) / legacy_years * & 
                 veget_max(ipts, ivm) / vegetmax_gp(ipts, ivmap)
               epidemic_monitor(ipts, ivm) = SUM(risk_index_legacy(ipts, ivm, :)) / legacy_years
             ENDDO ! iagec

             ! The start of an epidemic is only triggered by risk_index over the
             ! last legacy_years years. wght_sirdi_b(xvm) relate is a risk index
             ! value at wich wght_sirdi (the importance the rdi
             ! susceptibility )become dominant over woodleftover susceptibility
             ! this trick ensure a smooth transition from endemic to epidemic
             ! state. 
             IF(epidemic_risk(ipts, ivmap) > wght_sirdi_b(xvm)) THEN
               ! There is a minimum requierement to go out of an epidemic phase
               ! fist : the stand need to reach a RDI thresold meaning that the 
               ! there is not enough stems available for the actual beetle
               ! population which will fly away to another location.
               IF(rdi_gp(ipts,ivmap) .LT. rdi_limit(xvm)) THEN
                 DO iagec = 1, nagec_pft(ivmap)
                   IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                     CYCLE
                   ENDIF
                   ivm = start_index(ivmap) + (iagec - 1)
                   ! The fist time we go out of the epidemic, a lot of beetle
                   ! will flayaway to find new source of food. This reduce the
                   ! beetle pressure in the pixel and help the stand to come
                   ! back to a recover phase.
                   IF(epidemic(ipts, ivm) .EQ. 1) THEN
                     beetle_flyaway(ipts, ivm) = remaining_beetles(ivm)
                   ENDIF
                   epidemic(ipts, ivm) = 0
                 ENDDO
               ELSE
                 ! There is a bark beetle outbreak going on !
                 DO iagec = 1, nagec_pft(ivmap)
                   IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                     CYCLE
                   ENDIF
                   ivm = start_index(ivmap) + (iagec - 1)
                   epidemic(ipts, ivm) = 1
                 ENDDO    
               ENDIF           
             ! second : when the risk is low (<wght_sirdi_b), epidemic is
             ! impossible
             ELSE
               DO iagec = 1, nagec_pft(ivmap)
                 IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                   CYCLE
                 ENDIF
                 ivm = start_index(ivmap) + (iagec - 1)
                 epidemic(ipts, ivm) = 0
               ENDDO
             ENDIF

             DO iagec = 1, nagec_pft(ivmap)
               IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                 CYCLE
               ENDIF
               ivm = start_index(ivmap) + (iagec - 1)
               
               ! Beetle_flyaway alway come back to 1 (neutral) a rate define by
               ! legacy_years. 
               IF( beetle_flyaway(ipts,ivm) .LT. un ) THEN
                  IF(rdi_gp(ipts,ivmap) .LT. rdi_limit(xvm)) THEN
                    beetle_flyaway(ipts,ivm) = remaining_beetles(ivm)
                  ELSE
                    beetle_flyaway(ipts,ivm) = MIN(beetle_flyaway(ipts,ivm) + &
                      remaining_beetles(ivm)/(legacy_years*6),un)
                  ENDIF
               ENDIF

               ! This is where the importance of each specific susceptibilities
               ! are updated according to epidemic state and risk level. A S shape 
               ! relationship with risk level is used to ensure a smooth
               ! transition between epidemic and endemic states.
               IF( epidemic(ipts, ivm) .EQ. 1 ) THEN
                  !When an epidemic phase is triggered beetle feed on all the
                  !trees available. Beetles don't need woodleftover anymore.
                  wght_siw = zero
                  wght_sirdi = un - (wght_sis(xvm) + wght_sid(xvm) + wght_siw)
               ELSE
                  wght_sirdi = un  / ( un + exp(wght_sirdi_a(xvm) * (epidemic_risk(ipts, ivmap) - &
                        wght_sirdi_b(xvm)))) * (un - (wght_sis(xvm) + wght_sid(xvm)))
                  wght_siw = un - (wght_sirdi + wght_sis(xvm) + wght_sid(xvm))
               ENDIF

               ! Calculate the average bark beetle pressure. The average of three
               ! highest
               ! values from within the total number of legacy_years is used.
               ! bgi (number of beetle generation in one year)
               bgi_current(:) = beetle_generation_index(ipts, ivm, :)
 
               ! Sort the values to retrieve the three highest values.
               CALL quicksort(bgi_current, 1, legacy_years)
   
               ! Average the three recent years with the highest bark beetle
               ! pressure. The
               ! same bark beetle pressure is experienced by the whole pixel.
               ! G (number of beetle generation in one year)
               G(ipts, ivmap) = G(ipts, ivmap) + SUM(bgi_current(legacy_years - &
                        (nb_years_bgi - 1):legacy_years)) / &
                        nb_years_bgi * veget_max(ipts, ivm) / vegetmax_gp(ipts, ivmap)

               IF (.NOT. beetle_pft(ivm) .OR. veget_max(ipts, ivm) .LT. min_stomate) THEN
                   ! No reason to calculate bark beetle damage
                   CYCLE
               ELSE
                   IF(printlev_loc >= 4) THEN
                     WRITE(numout,*) 'bgi_current', bgi_current(:)
                     WRITE(numout,*) 'wood_leftover', wood_leftover_legacy(ipts, ivm, :)
                     WRITE(numout,*) 'beetle_damage_legacy', beetle_damage_legacy(ipts, ivm, :)
                     WRITE(numout,*) 'woody_frac', woody_frac(ipts,ivm)
                     WRITE(numout,*) 'wood_leftover_mean', &
                       SUM(wood_leftover_legacy(ipts, ivm, :)) / legacy_years - &
                       SUM(beetle_damage_legacy(ipts, ivm,:) * &
                       woody_frac(ipts,ivmap))
                     WRITE(numout,*) 'season_drought_mean',season_drought_mean(ipts, ivm)
                     WRITE(numout,*) 'epidemic_risk', epidemic_risk(ipts, ivmap)
                     WRITE(numout,*) 'epidemic flag', epidemic(ipts, ivm)
                     WRITE(numout,*) 'biomass_current', biomass_current(ipts,ivmap)
                     WRITE(numout,*) 'woody_current', woody_current(ipts, ivmap)
                     WRITE(numout,*) 'RDI group', rdi_gp(ipts, ivmap)
                     WRITE(numout,*) 'RDI', rdi(ipts, ivm)
                     WRITE(numout,*) 'wght_sirdi', wght_sirdi
                     WRITE(numout,*) 'wght_siw', wght_siw
                     WRITE(numout,*) 'beetle_flyaway', beetle_flyaway(ipts,ivm)
                     WRITE(numout,*) 'rdi_limit', rdi_limit(ivm) 
                   ENDIF


                   ! The self-thinning and yield relationships are for diameters
                   ! expressed in cm
                   rdi_suscept_gp(ipts, ivmap) = rdi_suscept_gp(ipts, ivmap) + &
                        (un / (un + exp(rdi_susceptibility_a(ivm) * &
                        (rdi(ipts, ivm) - rdi_susceptibility_b(ivm)))))* &
                        veget_max(ipts, ivm) / vegetmax_gp(ipts, ivmap)

                   !!!!!!! NOT USED ANYMORE !!!!!!!!
                   ! Stand age effect on bark beetle vulnerability. First calculaed at the
                   ! pft level and then accumulated for all age classes (thus accumulated
                   ! for the species group)
                   !standage_suscept(ipts, ivm) = age_susceptibility_a(ivm) + &
                   !     ( un - age_susceptibility_a(ivm)) / &
                   !     ( un + exp(-(age_susceptibility_b(ivm) * & 
                   !     (age_stand(ipts, ivm) - age_susceptibility_c(ivm)))))
                   !standage_suscept_gp(ipts, ivmap) = standage_suscept_gp(ipts, ivmap) + &
                   !     standage_suscept(ipts ,ivm) * & 
                   !     veget_max(ipts, ivm) / vegetmax_gp(ipts, ivmap)

                   ! LONG term drought susceptibility is an indirect measurement
                   ! tree defence to beetle colonisation. Multiple drought event
                   ! will decrease the amount of repelant produce by the tree.
                   drought_suscept(ipts,ivm) = un / (un + exp(drought_susceptibility_a(ivm) * &
                        ((un - season_drought_mean(ipts, ivm)) - drought_susceptibility_b(ivm))))
                   drought_suscept_gp(ipts, ivmap) = drought_suscept_gp(ipts, ivmap) + &
                        drought_suscept(ipts, ivm) * veget_max(ipts, ivm) / vegetmax_gp(ipts, ivmap)

                   ! Susceptibility_index is used to calculate the mortality
                   ! rate of infested trees. It is conceptually different from
                   ! susceptibility_index_group use to calculate trees that will
                   ! be infested by bark beetle. 
                   ! Mortality rate only consider susceptibilty
                   ! which affect physiological state of a tree like long term
                   ! drought or intraspecific comptition (RDI). windtrhow and
                   ! share are not present in the equation because rely on
                   ! mechanism involve only in the infesteation stage.   
                   suscept_index(ipts, ivm) = (un / (un + exp(rdi_susceptibility_a(ivm) * &
                        (rdi(ipts, ivm) - rdi_susceptibility_b(ivm))))) * wght_sirdi +  &
                        drought_suscept(ipts, ivm) * (un - wght_sirdi)
               ENDIF
             ENDDO ! end loop nagec

             ! The calculation of the windthrow suceptibility differs from the
             ! original paper of Temperli et al. 2013. First, the wood biomass
             ! lost by windthrow is replace by the wood leftover. This change
             ! implied that windthrow suceptibility is only representing beetle x
             ! windthrow interaction anymore. Of course, this index is still
             ! increasing drasticaly after a windthrow event. Second, the
             ! equation is using the actual woody biomass instead of the maximal
             ! potential woody biomass, a fix value of 300 t/ha in the paper, in
             ! ORCHIDEE we never reach this value and it will
             ! underestimate the contribution of windthrows.
             ! The amount of dead wood left on-site by previous wind storms is
             ! smaller than the living aboveground biomass. This will be accounted
             ! for to decrease the pft-level susceptibility to bark beetle attacks
             ! due to the left-over of windthrow debris. When 30% of
             ! the current wooody biomass is on the floor
             ! susceptibility reach its maximum(1)(unitless)
             windthrow_suscept_gp(ipts, ivmap) = MIN(un, (wood_leftover_mean(ipts, ivmap) / &
                        woody_current(ipts, ivmap)) / windthrow_susceptibility_tune(start_index(ivmap)))

             ! Calculate pixel sucseptibility at the level of the whole species group. It is
             ! assumed that half of the vulnerability comes from the relative density index. 
             ! But here RDI is not use has a proxy for competition but rather
             ! than a proxy for bettle spead rate wich is stimulatate when tree
             ! canopy close from each other. Long term drought (proxy of plant health) 
             ! and pixel deciduous share also account in the calculation. 
             ! wood leftover is also use to trigger the epidemic phase by increasing
             ! the risk index right after a storm but don't have any impact
             ! during the epidemic phase.
             ! susceptibility_index_group is multiplied with vegetmax to account for area in the
             ! pixel where susceptibility_index is zero because there is no suceptible forest
             ! i.e. picea abies. Finally susceptibility_index_group will be used
             ! to calculate the amount of picea tree that will be infested by
             ! bark beetle. 

             suscept_index_gp(ipts, ivmap) = &
                (drought_suscept_gp(ipts, ivmap) * wght_sid(ivm) + &
                 rdi_suscept_gp(ipts, ivmap) * wght_sirdi + &
                 share_suscept_gp(ipts, ivmap) * wght_sis(ivm) + &
                 windthrow_suscept_gp(ipts, ivmap) * wght_siw) * &
                vegetmax_gp(ipts, ivmap)

             DO iagec = 1, nagec_pft(ivmap)
               IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                 CYCLE
               ENDIF
               ivm = start_index(ivmap) + (iagec - 1)
               ! Estimate the risk index at the Age class level.
               suscept_index_monitor(ipts, ivm) = suscept_index_gp(ipts, ivmap)
               windthrow_suscept_monitor(ipts, ivm) = windthrow_suscept_gp(ipts,ivmap)
                
             END DO


             !Debug
             IF(printlev_loc >= 4) THEN
                WRITE(numout,*) 'drought_susceptibility', drought_suscept_gp(ipts, ivmap)
                WRITE(numout,*) 'rdi_supceptibilit', rdi_suscept_gp(ipts, ivmap)
                WRITE(numout,*) 'share_susceptibility', share_suscept_gp(ipts, ivmap)
                WRITE(numout,*) 'windthrow_susceptibility', windthrow_suscept_gp(ipts, ivmap)
             END IF
             !-
             
             ! Calculate as the real pressure of the beetles. Avegare of the average bark
             ! beetle of the past three years and the beetle legacy (which is a ratio between
             ! affected dead wood (breeding grounds) and fresh biomass (food)). This average
             ! is an index for the available beetles when multiplied with the susceptibility
             ! of the species group we get the beetle pressure (which could be seen as a proxy
             ! for beetle biomass). Make sure the index stays between zero and one. Here the
             ! lower bound was set to 0,05 to reflect the fact that there is always an endemic
             ! population of bark beetles. If we would allow this index to go to zero beetles
             ! could go extint. Given that both G and beetle_legacy_index can reach zero it is
             ! slightly easier to control the value of the beetle_pressure_index here. 
             beetle_pressure_index(ipts, ivmap) = MIN(MAX(suscept_index_gp(ipts, ivmap) * &
                  (G(ipts, ivmap) + beetle_pop_legacy(ipts, xvm, 1)) * 0.5* &
                  beetle_flyaway(ipts, ivmap), 0.05), un)

             ! Estimate the risk index at the pixel level.
             risk_index_gp(ipts, ivmap) =  suscept_index_gp(ipts, ivmap) * &
                  beetle_pressure_index(ipts, ivmap)

             ! This loop is use for monitoring intermediate result.
             ! ivmap ---> ivm, because there is no ivmap dimension in the
             ! output files 
             DO iagec = 1, nagec_pft(ivmap)
               IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                 CYCLE
               ENDIF
               ivm = start_index(ivmap) + (iagec - 1)
               risk_index(ipts, ivm) = risk_index_gp(ipts, ivmap)
               suscept_index_monitor(ipts, ivm) = suscept_index_gp(ipts, ivmap)
               beetle_pressure_monitor(ipts, ivm) = beetle_pressure_index(ipts, ivmap)
               windthrow_suscept_monitor(ipts, ivm) = windthrow_suscept_gp(ipts, ivmap)
             END DO

             !Debug
             IF(printlev_loc >= 4) THEN
                WRITE(numout,*) 'G', G(ipts, ivmap)
                WRITE(numout,*) 'beetle_pressure_index', beetle_pressure_index(ipts, ivmap)
                WRITE(numout,*) 'beetle_pop_legacy', beetle_pop_legacy(ipts, xvm, 1)
                WRITE(numout,*) 'risk_index_group', risk_index_gp(ipts, :)
                WRITE(numout,*) 'susceptibility_index_group', suscept_index_gp(ipts, ivmap)
             END IF
             !-
                
             ! Estimate the total biomass of spruce infected by bark beetle. The
             ! calculation diverge from the original equation which used a fix
             ! value corresponding the half of the maximun potential biomass (also fix).
             ! In our case maximun potential biomass is not a fix value and depend
             ! climate forcing, soil properties, interspecies water competition. Seems
             ! easier to use the actual biomass and recalibrate suceptibility parameters
             ! in order to decrease the overall risk index values.
             biomass_infected(ipts, ivmap) = biomass_current(ipts, ivmap) * risk_index_gp(ipts, ivmap)

             ! Calculate Beetle pressure index. The actual biomass(biomass_current) is
             ! used instead of the maximal potential biomass, i.e., 300 t/ha, as was done
             ! in temperli et al. 2013. This approach differs from Temperli. The ratio
             ! between dead wood from previous beetle outbreaks (breeding grounds) and
             ! uninfected biomass (fresh food) is seen as a driver of new infections.
             ! Make sure the legacy stays inbetween zero and one.
             beetle_temp1(ipts, ivm, :) = zero

             DO iyear = 1, legacy_years - 1
               beetle_temp1(ipts, ivm, iyear + 1) = beetle_pop_legacy(ipts, ivm, iyear)
             ENDDO

             beetle_temp1(ipts, ivm, 1) = MIN(biomass_infected(ipts, ivmap) / &
                (biomass_current(ipts, ivmap) * pressure_feedback(ivm)), un)
             beetle_pop_legacy(ipts, ivm, :) = beetle_temp1(ipts, ivm, :)

             ! Distribute the total infected biomass over the different age classes
             ! of the same species
             DO iagec = 1, nagec_pft(ivmap)
                IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                  CYCLE
                ENDIF
                ! Determine the PFT index for the age classes within the species
                ! group under consideration
                ivm = start_index(ivmap) + (iagec - 1)

                ! Estimate mortality rate at the PFT level (unitless).
                mortality_rate(ipts, ivm) = (suscept_index(ipts, ivm) + &
                     beetle_pressure_index(ipts, ivmap)) * 0.5

                ! Estimate living biomass killed by bark beetles (gC or gN m-2)
                beetle_damage(ipts, ivm) = mortality_rate(ipts, ivm) * &
                     biomass_infected(ipts, ivmap) * &
                     veget_max(ipts, ivm) / &
                     vegetmax_gp(ipts, ivmap)

                beetle_temp2(ipts, ivm, :) = zero
                DO iyear = 1, beetle_legacy - 1
                  beetle_temp2(ipts, ivm, iyear + 1) = & 
                     beetle_damage_legacy(ipts, ivm, iyear)
                ENDDO 
                beetle_temp2(ipts, ivm, 1) = beetle_damage(ipts, ivm)
                beetle_damage_legacy(ipts, ivm, :) = beetle_temp2(ipts, ivm, :)

                !Debug
                IF(printlev_loc >= 4) THEN
                   WRITE(numout,*) 'biomass_current', biomass_current(ipts,ivmap)
                   WRITE(numout,*) 'biomass_infected', biomass_infected(ipts, ivmap)
                   WRITE(numout,*) 'beetle_damage', beetle_damage(ipts, ivm)
                   WRITE(numout,*) 'effective mortality in %', beetle_damage(ipts, ivm) / &
                     biomass_current(ipts,ivmap) * 100
                   WRITE(numout,*) 'mortality_rate', mortality_rate(ipts, ivm)
                   WRITE(numout,*) 'vegetmax_group', vegetmax_gp(ipts, ivmap)
                   WRITE(numout,*) 'veget_max', veget_max(ipts, ivm)
                END IF
                !-
                ! Select which circ_class would be damaged based on the thier
                ! diameters
                ncirc_loc = 0
                qdia(:) = wood_to_dia(circ_class_biomass(ipts, ivm, :, :, icarbon), ivm)
                DO icir = 1, ncirc
                  IF (qdia(icir) > 0.075) THEN
                    ncirc_loc = ncirc_loc + 1
                    WRITE(numout,*) 'diameter in cm:', qdia(icir) * 100
                  ENDIF
                ENDDO
                WRITE(numout,*) 'ncirc_loc: ', ncirc_loc
                ! By using an decreasing index, I made the assumption that the
                ! first
                ! tree which die from bark beetle infection are the bigest one.
                DO icir = ncirc_loc, 1, -1
                   IF(SUM(circ_class_biomass(ipts, ivm, icir, :, icarbon)) * &
                        circ_class_n(ipts, ivm, icir) .LT. beetle_damage(ipts, ivm)) THEN
                      ! All individuals in the circumference class need to be killed to satisfy
                      ! the total amount of biomass that needs to be killed.
                      circ_class_kill(ipts, ivm, icir, forest_managed(ipts, ivm), icut_beetle) = &
                           circ_class_n(ipts, ivm, icir)
                      beetle_damage(ipts, ivm) = beetle_damage(ipts, ivm) - &
                           SUM(circ_class_biomass(ipts,ivm, icir, :, icarbon)) * circ_class_n(ipts, ivm, icir)
                   ELSE
                      IF(beetle_damage(ipts, ivm) .GT. zero) THEN
                         ! Only part of the trees in this circumference class should be killed
                         circ_class_kill(ipts, ivm, icir, forest_managed(ipts, ivm), icut_beetle) = &
                              circ_class_n(ipts, ivm, icir) * beetle_damage(ipts, ivm) / &
                              (SUM(circ_class_biomass(ipts, ivm, icir, :, icarbon)) * circ_class_n(ipts, ivm, icir))
                         beetle_damage(ipts, ivm) = zero
                      ELSE
                         ! The total demand of mortality has been satisfied. No more killing is
                         ! required.
                         circ_class_kill(ipts, ivm, icir, forest_managed(ipts, ivm), icut_beetle) = zero
                      ENDIF
                   ENDIF
                ENDDO ! End loop ncirc

             ENDDO   ! End loop nagec

          ELSE
             ! No mortality from bark beetles. Set the mortality to zero to avoid problems
             ! with uninitialized values
             DO iagec = 1, nagec_pft(ivmap)
                IF (nc(ipts, ivmap, iagec) .EQ. 0) THEN
                  CYCLE
                ENDIF
                ! Determine the PFT index for the age classes within the species
                ! group under consideration
                ivm = start_index(ivmap) + (iagec - 1)
                ! No trees are to be killed
                circ_class_kill(ipts, ivm, :, forest_managed(ipts, ivm), icut_beetle) = zero
             ENDDO ! nagec
          ENDIF ! beetle_group .EQV. .TRUE. .AND. vegetmax_group(ipts,ivmap) .GT. min_stomate

       ENDDO ! End loop nvmap     
    ENDDO ! End loop npts

  END SUBROUTINE bark_beetle_damage


! +++MOVE TO FUNCTION_LIBRARY.F90
! +++COMPLETE HEADER  
!=============================================================================================================================
!! SUBROUTINE    : quicksort
!!
!>\BRIEF
!!
!!\n DESCRIPTION : 
!!  
!! RECENT CHANGE(S): 
!!
!! MAIN OUTPUT VARIABLE(S) : 
!!
!! REFERENCES: 
!!
!! FLOWCHART     : 
!! \n
!_
!=============================================================================================================================

  ! quicksort.f -*-f90-*-
  ! Author: t-nissie
  ! License: GPLv3
  ! Gist: https://gist.github.com/t-nissie/479f0f16966925fa29ea
  RECURSIVE SUBROUTINE quicksort(a, first, last)
    IMPLICIT NONE
    REAL*8  a(*), x, t
    INTEGER first, last
    INTEGER i, j
!_
!=============================================================================================================================
    x = a( (first + last) * 0.5 )
    i = first
    j = last
    DO
      DO WHILE (a(i) < x)
        i = i + 1
     ENDDO
      DO WHILE (x < a(j))
        j = j - 1
      ENDDO
      IF (i >= j) EXIT
      t = a(i);  a(i) = a(j);  a(j) = t
      i = i + 1
      j = j - 1
    ENDDO
    IF (first < i - 1) call quicksort(a, first, i - 1)
    IF (j + 1 < last)  call quicksort(a, j + 1, last)
  END SUBROUTINE quicksort
  
END MODULE stomate_pest

